import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='wien2k_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='wien2k.nomadmetainfo.json'))


class x_wien2k_header(MSection):
    '''
    header (labels) of wien2k.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_wien2k_header'))

    x_wien2k_release_date = Quantity(
        type=str,
        shape=[],
        description='''
        Release date of wien2k.
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_release_date'))

    x_wien2k_version = Quantity(
        type=str,
        shape=[],
        description='''
        Version of WIEN2k.
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_version'))


class x_wien2k_section_XC(MSection):
    '''
    exchange-correlation potential, in in0
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_wien2k_section_XC'))

    x_wien2k_indxc = Quantity(
        type=str,
        shape=[],
        description='''
        exchange-correlation potential, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_indxc'))


class x_wien2k_section_equiv_atoms(MSection):
    '''
    section containing a class of equivalent atoms
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_wien2k_section_equiv_atoms'))

    x_wien2k_atom_pos_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom x in internal units
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atom_pos_x'))

    x_wien2k_atom_pos_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom y in internal units
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atom_pos_y'))

    x_wien2k_atom_pos_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom z  in internal units
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atom_pos_z'))

    x_wien2k_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom, labelling non-equvalent atoms
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atom_name'))

    x_wien2k_NPT = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of radial mesh points
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_NPT'))

    x_wien2k_RMT = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic sphere radius (muffin-tin radius)
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_RMT'))

    x_wien2k_R0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        first radial mesh point
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_R0'))

    x_wien2k_atomic_number_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic number Z
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atomic_number_Z'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_wien2k_header = SubSection(
        sub_section=SectionProxy('x_wien2k_header'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_wien2k_header'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_wien2k_iteration_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        scf iteration number
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_iteration_number'))

    x_wien2k_nr_of_independent_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of independent atoms in the cell
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_nr_of_independent_atoms'))

    x_wien2k_potential_option = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        exchange correlation potential option
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_potential_option'))

    x_wien2k_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_system_name'))

    x_wien2k_total_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        total number of atoms in the cell
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_total_atoms'))

    x_wien2k_lattice_const_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter a in this calculation
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_lattice_const_a'))

    x_wien2k_lattice_const_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter b in this calculation
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_lattice_const_b'))

    x_wien2k_lattice_const_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter c in this calculation
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_lattice_const_c'))

    x_wien2k_unit_cell_volume_bohr3 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='bohr ** 3',
        description='''
        unit cell volume
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_unit_cell_volume_bohr3'))

    x_wien2k_energy_total = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        energy total
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_energy_total'))

    x_wien2k_fermi_ene = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Fermi energy
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_fermi_ene'))

    x_wien2k_noe = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of electrons
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_noe'))

    x_wien2k_nr_kpts = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of k-points
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_nr_kpts'))

    x_wien2k_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Potential and charge cut-off, Ry**.5
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_cutoff'))

    x_wien2k_ene_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        energy gap in Ry
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ene_gap'))

    x_wien2k_ene_gap_eV = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='electron_volt',
        description='''
        energy gap in eV
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ene_gap_eV'))

    x_wien2k_matrix_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        matrix size
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_matrix_size'))

    x_wien2k_rkm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rkm
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_rkm'))

    x_wien2k_LOs = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        LOs
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_LOs'))

    x_wien2k_mmtot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total magnetic moment in cell
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_mmtot'))

    x_wien2k_mmint = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        magnetic moment in the interstital region
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_mmint'))

    x_wien2k_mmi001 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        magnetic moment inside the sphere
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_mmi001'))

    x_wien2k_charge_distance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge distance between last 2 iterations
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_charge_distance'))

    x_wien2k_for_abs = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): |F|
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_abs'))

    x_wien2k_for_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fx
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_x'))

    x_wien2k_for_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fy
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_y'))

    x_wien2k_for_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fz
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_z'))

    x_wien2k_for_x_gl = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fx
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_x_gl'))

    x_wien2k_for_y_gl = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fy
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_y_gl'))

    x_wien2k_for_z_gl = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fz
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_for_z_gl'))

    x_wien2k_atom_nr = Quantity(
        type=str,
        shape=[],
        description='''
        number of atom, labelling atoms
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_atom_nr'))

    x_wien2k_sphere_nr = Quantity(
        type=str,
        shape=[],
        description='''
        number of sphere, labelling spheres
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_sphere_nr'))

    x_wien2k_tot_diff_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total difference charge density for atom xx between last 2 iterations
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_diff_charge'))

    x_wien2k_tot_int_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total interstitial charge (mixed after MIXER)
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_int_charge'))

    x_wien2k_tot_charge_in_sphere = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total charge in sphere xx (mixed after MIXER)
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_charge_in_sphere'))

    x_wien2k_tot_int_charge_nm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total interstitial charge (new (not mixed) from LAPW2+LCORE
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_int_charge_nm'))

    x_wien2k_tot_charge_in_sphere_nm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total charge in sphere xx (new (not mixed) from LAPW2+LCORE
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_charge_in_sphere_nm'))

    x_wien2k_tot_val_charge_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total valence charge inside unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_val_charge_cell'))

    x_wien2k_tot_val_charge_sphere = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total valence charge in sphere xx
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_tot_val_charge_sphere'))

    x_wien2k_density_at_nucleus_valence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density for atom xx at the nucleus (first radial mesh point); valence
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_density_at_nucleus_valence'))

    x_wien2k_density_at_nucleus_semicore = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density for atom xx at the nucleus (first radial mesh point); semi-core
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_density_at_nucleus_semicore'))

    x_wien2k_density_at_nucleus_core = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density for atom xx at the nucleus (first radial mesh point); core
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_density_at_nucleus_core'))

    x_wien2k_density_at_nucleus_tot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        density for atom xx at the nucleus (first radial mesh point); total
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_density_at_nucleus_tot'))

    x_wien2k_nuclear_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        nuclear and electronic charge; normalization check of electronic charge densities.
        If a significant amount of electrons is missing, one might have core states, whose
        charge density is not completely confined within the respective atomic sphere. In
        such a case the corresponding states should be treated as band states (using LOs).
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_nuclear_charge'))

    x_wien2k_electronic_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        nuclear and electronic charge; normalization check of electronic charge densities.
        If a significant amount of electrons is missing, one might have core states, whose
        charge density is not completely confined within the respective atomic sphere. In
        such a case the corresponding states should be treated as band states (using LOs).
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_electronic_charge'))

    x_wien2k_necnr = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of the nec test, labelling nec
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_necnr'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_wien2k_nonequiv_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of inequivalent atoms in the unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_nonequiv_atoms'))

    x_wien2k_system_nameIn = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system given in the struct file
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_system_nameIn'))

    x_wien2k_calc_mode = Quantity(
        type=str,
        shape=[],
        description='''
        relativistic or nonrelativistic calculation mode
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_calc_mode'))

    x_wien2k_unit_cell_param_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - a
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_unit_cell_param_a'))

    x_wien2k_unit_cell_param_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - b
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_unit_cell_param_b'))

    x_wien2k_unit_cell_param_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_unit_cell_param_c'))

    x_wien2k_angle_between_unit_axis_alfa = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - alfa
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_angle_between_unit_axis_alfa'))

    x_wien2k_angle_between_unit_axis_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - beta
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_angle_between_unit_axis_beta'))

    x_wien2k_angle_between_unit_axis_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - gamma
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_angle_between_unit_axis_gamma'))

    x_wien2k_section_equiv_atoms = SubSection(
        sub_section=SectionProxy('x_wien2k_section_equiv_atoms'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_wien2k_section_equiv_atoms'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_wien2k_switch = Quantity(
        type=str,
        shape=[],
        description='''
        switch in in0 between TOT, KXC, POT, MULT, COUL, EXCH
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_switch'))

    x_wien2k_ifft_x = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in x direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ifft_x'))

    x_wien2k_ifft_y = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in y direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ifft_y'))

    x_wien2k_ifft_z = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in z direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ifft_z'))

    x_wien2k_ifft_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Multiplicative factor to the IFFT grid, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_ifft_factor'))

    x_wien2k_iprint = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        optional print switch, in in0
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_iprint'))

    x_wien2k_wf_switch = Quantity(
        type=str,
        shape=[],
        description='''
        wave function switch between WFFIL, SUPWF, WPPRI
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_wf_switch'))

    x_wien2k_rkmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RmtKmax - determines matrix size (convergence), where Kmax is the plane wave cut-
        off, Rmt is the smallest of all atomic sphere radii
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_rkmax'))

    x_wien2k_in2c_switch = Quantity(
        type=str,
        shape=[],
        description='''
        switch, in in2c between (TOT,FOR,QTL,EFG,ALM,CLM,FERMI)
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_switch'))

    x_wien2k_in2c_emin = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lower energy cut-off for defining the range of occupied states; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_emin'))

    x_wien2k_in2c_ne = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of electrons (per unit cell) in given energy range in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_ne'))

    x_wien2k_in2c_espermin = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        LAPW2 tries to find the .mean. energies for each l channel, for both the valence
        and the semicore states. To define .valence. and .semicore. it starts at (EF -
        .esepermin.) and searches for a .gap. with a width of at least .eseper0. and
        defines this as separation energy of valence and semicore; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_espermin'))

    x_wien2k_in2c_esper0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        minimum gap width; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_esper0'))

    x_wien2k_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        determines how EF is determined; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_smearing_kind'))

    x_wien2k_smearing_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        specifies the width of the broadening, if efmod is set to ALL, eval specifies the
        upper limit of the energy window; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_smearing_width'))

    x_wien2k_in2c_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        max. G (magnitude of largest vector) in charge density Fourier expansion; in in2c
        ''',
        a_legacy=LegacyDefinition(name='x_wien2k_in2c_gmax'))

    x_wien2k_section_XC = SubSection(
        sub_section=SectionProxy('x_wien2k_section_XC'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_wien2k_section_XC'))


m_package.__init_metainfo__()



from typing import Union

from fastapi import APIRouter, Depends, Request

from optimade.models import (
    ErrorResponse,
    StructureResource,
    StructureResponseMany,
    StructureResponseOne,
)
from optimade.server.config import CONFIG
from optimade.server.entry_collections import MongoCollection, client
from optimade.server.mappers import StructureMapper
from optimade.server.query_params import EntryListingQueryParams, SingleEntryQueryParams

from .utils import get_entries, get_single_entry

router = APIRouter(redirect_slashes=True)

structures_coll = MongoCollection(
    collection=client[CONFIG.mongo_database][CONFIG.structures_collection],
    resource_cls=StructureResource,
    resource_mapper=StructureMapper,
)


@router.get(
    "/structures",
    response_model=Union[StructureResponseMany, ErrorResponse],
    response_model_exclude_unset=True,
    tags=["Structures"],
)
def get_structures(request: Request, params: EntryListingQueryParams = Depends()):
    return get_entries(
        collection=structures_coll,
        response=StructureResponseMany,
        request=request,
        params=params,
    )


@router.get(
    "/structures/{entry_id:path}",
    response_model=Union[StructureResponseOne, ErrorResponse],
    response_model_exclude_unset=True,
    tags=["Structures"],
)
def get_single_structure(
    request: Request, entry_id: str, params: SingleEntryQueryParams = Depends()
):
    return get_single_entry(
        collection=structures_coll,
        entry_id=entry_id,
        response=StructureResponseOne,
        request=request,
        params=params,
    )



# Copyright 2016-2018 Martina Stella, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from .OnetepCommon import get_metaInfo
import logging, os, re, sys



############################################################
# This is the parser for the *.cell file of Onetep.
############################################################

logger = logging.getLogger("nomad.OnetepCellParser")

class OnetepCellParserContext(object):
    """Context for parsing Onetep *.cell file.


    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of wrting and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData
        self.cell_store = []
        self.at_nr = []
        self.onetep_atom_positions_store=[]
        self.atom_labels_store=[]

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser

        # get unit from metadata for band energies
        # allows to reset values if the same superContext is used to parse different files
        # self.band_energies = None
        # self.band_k_points = None
        # self.band_occupations = None

        # self.k_crd = []
        # self.k_sgt_start_end = []
    def onClose_x_onetep_section_cell(self, backend, gIndex, section):
        """trigger called when _onetep_section_cell is closed"""
        # get cached values for onetep_cell_vector
        units = section['x_onetep_units']
        vet = section['x_onetep_cell_vector']
        bohr_to_m = float(5.29177211e-11)
        ang_to_m = float(1e-10)
        if units:
            if units[0] == 'bohr':
                vet[0] = vet[0].split()
                vet[0] = [float(i) for i in vet[0]]
                vet[0]= [i * bohr_to_m for i in vet[0]]
                vet[1] = vet[1].split()
                vet[1] = [float(i) for i in vet[1]]
                vet[1]= [i * bohr_to_m for i in vet[1]]
                vet[2] = vet[2].split()
                vet[2] = [float(i) for i in vet[2]]
                vet[2]= [i * bohr_to_m for i in vet[2]]
                self.cell_store.append(vet[0])
                self.cell_store.append(vet[1])
                self.cell_store.append(vet[2]) # Reconstructing the unit cell vector by vector
            elif units[0] == 'ang':
                vet[0] = vet[0].split()
                vet[0] = [float(i) for i in vet[0]]
                vet[0]= [i * ang_to_m for i in vet[0]]
                vet[1] = vet[1].split()
                vet[1] = [float(i) for i in vet[1]]
                vet[1]= [i * ang_to_m for i in vet[1]]
                vet[2] = vet[2].split()
                vet[2] = [float(i) for i in vet[2]]
                vet[2]= [i * ang_to_m for i in vet[2]]
                self.cell_store.append(vet[0])
                self.cell_store.append(vet[1])
                self.cell_store.append(vet[2]) # Reconstructing t
        else:
            vet[0] = vet[0].split()
            vet[0] = [float(i) for i in vet[0]]
            vet[0]= [i * bohr_to_m for i in vet[0]]
            vet[1] = vet[1].split()
            vet[1] = [float(i) for i in vet[1]]
            vet[1]= [i * bohr_to_m for i in vet[1]]
            vet[2] = vet[2].split()
            vet[2] = [float(i) for i in vet[2]]
            vet[2]= [i * bohr_to_m for i in vet[2]]
            self.cell_store.append(vet[0])
            self.cell_store.append(vet[1])
            self.cell_store.append(vet[2]) # Reco

    def onClose_section_system(self, backend, gIndex, section):
        unitsap = section['x_onetep_units_atom_position']
        pos = section['x_onetep_store_atom_positions']

        bohr_to_m = float(5.29177211e-11)
        ang_to_m = float(1e-10)
        if unitsap is not None:
            if unitsap[0] == 'bohr':

                pos = section['x_onetep_store_atom_positions']

                if pos:
                    self.at_nr = len(pos)
                    for i in range(0, self.at_nr):
                        pos[i] = pos[i].split()
                        pos[i] = [float(j) for j in pos[i]]
                        pos[i]= [ii * bohr_to_m for ii in pos[i]]
                        self.onetep_atom_positions_store.append(pos[i])

            elif unitsap[0] == 'ang':
                pos = section['x_onetep_store_atom_positions']

                if pos:
                    self.at_nr = len(pos)
                    for i in range(0, self.at_nr):
                        pos[i] = pos[i].split()
                        pos[i] = [float(j) for j in pos[i]]
                        pos[i]= [ii * ang_to_m for ii in pos[i]]

                        self.onetep_atom_positions_store.append(pos[i])
            else:
                pass





        if pos:
            bohr_to_m = float(5.29177211e-11)
            self.at_nr = len(pos)
            for i in range(0, self.at_nr):
                pos[i] = pos[i].split()
                pos[i] = [float(j) for j in pos[i]]
                pos[i]= [ii * bohr_to_m for ii in pos[i]]
                self.onetep_atom_positions_store.append(pos[i])
        #get cached values of onetep_store_atom_labels

        lab = section['x_onetep_store_atom_labels']

        for i in range(0, len(lab)):
            lab[i] = re.sub('\s+', ' ', lab[i]).strip()
        self.atom_labels_store.extend(lab)



def build_OnetepCellFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.cell file of Onetep.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.cell file of Onetep.
    """

    return SM(name = "systemDescription",
                startReStr = "",
                forwardMatch = True,
                sections = ["section_system"],
                subMatchers = [

           # cell information
                    SM(name = 'cellInformation',
                        startReStr = r"\%block\slattice\_cart\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onetep_section_cell
                    SM(name = 'cellInformation',
                        startReStr = r"\s*\%block\slattice\_cart\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onetep_section_cell
                    SM(name = 'cellInformation',
                        startReStr = r"\%block\s*lattice\_cart\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onet
                    SM(name = 'cellInformation',
                        startReStr = r"\s*\%block\s*lattice\_cart\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onet

                    SM(name = 'cellInformation',
                        startReStr = r"\%BLOCK\sLATTICE\_CART\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onetep_section_cell
                    SM(name = 'cellInformation',
                        startReStr = r"\s*\%BLOCK\sLATTICE\_CART\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onetep_section_cell
                    SM(name = 'cellInformation',
                        startReStr = r"\%BLOCK\s*LATTICE\_CART\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onet
                    SM(name = 'cellInformation',
                        startReStr = r"\s*\%BLOCK\s*LATTICE\_CART\s*",
                        # forwardMatch = True,
                        sections = ["x_onetep_section_cell"],
                            subMatchers = [
                                SM(r"(?P<x_onetep_units>[A-Za-z]+)"),
                                SM(r"\s*(?P<x_onetep_cell_vector>[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEd]+)",
                     #SM(r"\s*(?P<onetep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                        # endReStr = "\%endblock\s*lattice\_cart\s*",
                                repeats = True),

                            ]), # CLOSING onet

                    SM(startReStr = r"\%block\spositions\_abs\s*",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),

                        SM(r"(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOSING onet
                    SM(startReStr = r"\s*\%block\spositions\_abs\s*",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"\s*(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOS
                    SM(startReStr = r"\%block\s*positions\_abs\s*",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOSING onetep_section_atom_position
                    SM(startReStr = r"\s*\%block\s*positions\_abs\s*",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"\s*(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOSI



                    SM(startReStr = r"\%BLOCK\sPOSITIONS\_ABS",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOSING onet
                    SM(startReStr = r"\s*\%BLOCK\sPOSITIONS\_ABS",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"\s*(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOS
                    SM(startReStr = r"\%BLOCK\s*POSITIONS\_ABS",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # endReStr = "\n",
                            repeats = True)

                                 ]), # CLOSING onetep_section_atom_position
                    SM(startReStr = r"\s*\%BLOCK\s*POSITIONS\_ABS",
                        forwardMatch = True,
                        sections = ["x_onetep_section_atom_positions"],
                        endReStr = "\n",
                        subMatchers = [
                        SM(r"(?P<x_onetep_units_atom_position>[A-Za-z]+)"),
                        SM(r"\s*(?P<x_onetep_store_atom_labels>[A-Za-z0-9]+)\s*(?P<x_onetep_store_atom_positions>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)",
                            # ,
                            repeats = True)

                                 ]), # CLOSI
                   ]) # CLOSING SM systemDescription





def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_system': CachingLvl,
                                # 'section_run': CachingLvl,
                                # 'section_single_configuration_calculation': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_onetep_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the Onetep *.cell file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    OnetepCellFileSimpleMatcher = build_OnetepCellFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/Onetep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/onetep.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'Onetep-cell-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = OnetepCellFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = OnetepCellParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)










# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module comprises the nomad@FAIRDI APIs. Currently there is NOMAD's official api, and
we will soon at the optimade api. The app module also servers documentation, gui, and
alive.
'''
from flask import Flask, Blueprint, jsonify, url_for, abort, request, make_response
from flask_restplus import Api, representations
from flask_cors import CORS
from werkzeug.exceptions import HTTPException
from werkzeug.wsgi import DispatcherMiddleware  # pylint: disable=E0611
import os.path
import random
from structlog import BoundLogger
import collections
from mongoengine.base.datastructures import BaseList
import orjson

from nomad import config, utils as nomad_utils

from .api import blueprint as api_blueprint, api
from .optimade import blueprint as optimade_blueprint, api as optimade
from .docs import blueprint as docs_blueprint
from .dist import blueprint as dist_blueprint
from .gui import blueprint as gui_blueprint
from .encyclopedia import blueprint as encyclopedia_blueprint
from . import common


def dump_json(data):
    def default(data):
        if isinstance(data, collections.OrderedDict):
            return dict(data)

        if data.__class__.__name__ == 'BaseList':
            return list(data)

        raise TypeError

    return orjson.dumps(
        data, default=default,
        option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS)


# replace the json implementation of flask_restplus
def output_json(data, code, headers=None):
    dumped = dump_json(data) + b'\n'

    resp = make_response(dumped, code)
    resp.headers.extend(headers or {})
    return resp


api.representation('application/json')(output_json)
optimade.representation('application/json')(output_json)


@property  # type: ignore
def specs_url(self):
    '''
    Fixes issue where swagger-ui makes a call to swagger.json over HTTP.
    This can ONLY be used on servers that actually use HTTPS.  On servers that use HTTP,
    this code should not be used at all.
    '''
    return url_for(self.endpoint('specs'), _external=True, _scheme='https')


if config.services.https:
    Api.specs_url = specs_url


app = Flask(__name__)
''' The Flask app that serves all APIs. '''

app.config.APPLICATION_ROOT = common.base_path  # type: ignore
app.config.RESTPLUS_MASK_HEADER = False  # type: ignore
app.config.RESTPLUS_MASK_SWAGGER = False  # type: ignore
app.config.SWAGGER_UI_OPERATION_ID = True  # type: ignore
app.config.SWAGGER_UI_REQUEST_DURATION = True  # type: ignore

app.config['SECRET_KEY'] = config.services.api_secret


def api_base_path_response(env, resp):
    resp('200 OK', [('Content-Type', 'text/plain')])
    return [
        ('Development nomad api server. Api is served under %s/.' %
            config.services.api_base_path).encode('utf-8')]


app.wsgi_app = DispatcherMiddleware(  # type: ignore
    api_base_path_response, {config.services.api_base_path: app.wsgi_app})

CORS(app)

app.register_blueprint(api_blueprint, url_prefix='/api')
app.register_blueprint(optimade_blueprint, url_prefix='/optimade')
app.register_blueprint(docs_blueprint, url_prefix='/docs')
app.register_blueprint(dist_blueprint, url_prefix='/dist')
app.register_blueprint(gui_blueprint, url_prefix='/gui')
app.register_blueprint(encyclopedia_blueprint, url_prefix='/encyclopedia')


@app.errorhandler(Exception)
def handle(error: Exception):
    status_code = getattr(error, 'code', 500)
    if not isinstance(status_code, int):
        status_code = 500
    if status_code < 100:
        status_code = 500

    name = getattr(error, 'name', 'Internal Server Error')
    description = getattr(error, 'description', 'No description available')
    data = dict(
        code=status_code,
        name=name,
        description=description)
    data.update(getattr(error, 'data', []))
    response = jsonify(data)
    response.status_code = status_code
    if status_code == 500:
        local_logger = common.logger
        # the logger is created in before_request, if the error was created before that
        # logger can be None
        if local_logger is None:
            local_logger = nomad_utils.get_logger(__name__)

        # TODO the error seems not to be the actual exception, therefore
        # there might be no stacktrace. Maybe there is a way to get the actual
        # exception/stacktrace
        local_logger.error('internal server error', error=str(error), exc_info=error)

    return response


@app.route('/alive')
def alive():
    ''' Simple endpoint to utilize kubernetes liveness/readiness probing. '''
    return "I am, alive!"


@app.before_request
def before_request():
    # api logger
    args = getattr(request, 'view_args')
    if args is None:
        args = {}
    else:
        args = dict(**args)

    args.update(
        name=__name__,
        blueprint=str(request.blueprint),
        endpoint=request.endpoint,
        method=request.method,
        url=request.url,
        json=request.json,
        args=request.args)

    common.logger = nomad_utils.get_logger(**args)

    # chaos monkey
    if config.services.api_chaos > 0:
        if random.randint(0, 100) <= config.services.api_chaos:
            abort(random.choice([400, 404, 500]), 'With best wishes from the chaos monkey.')


@app.before_first_request
def setup():
    from nomad import infrastructure

    if not app.config['TESTING']:
        # each subprocess is supposed disconnect connect again: https://jira.mongodb.org/browse/PYTHON-2090
        try:
            from mongoengine import disconnect
            disconnect()
        except Exception:
            pass

        infrastructure.setup()



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# from builtins import next
# from builtins import range
from builtins import next
from builtins import range
import os
import re
import logging
import importlib
from nomadcore.baseclasses import ParserInterface
logger = logging.getLogger("nomad")


class NWChemParser(ParserInterface):
    """This class handles the initial setup before any parsing can happen. It
    determines which version of NWChem was used to generate the output and then
    sets up a correct main parser.

    After the implementation has been setup, you can parse the files with
    parse().
    """
    def __init__(self, metainfo_to_keep=None, backend=None, default_units=None, metainfo_units=None, debug=True, log_level=logging.ERROR, store=True):
        super(NWChemParser, self).__init__(metainfo_to_keep, backend, default_units, metainfo_units, debug, log_level, store)

    def setup_version(self):
        """Setups the version by looking at the output file and the version
        specified in it.
        """
        # Search for the NWChem version specification. The correct parser is
        # initialized based on this information.
        regex_version = re.compile("\s+Northwest Computational Chemistry Package \(NWChem\) (\d+\.\d+(?:\.\d+)?)")
        version_id = None
        with open(self.parser_context.main_file, 'r') as outputfile:
            for line in outputfile:
                # Look for version
                result_version = regex_version.match(line)
                if result_version:
                    version_id = result_version.group(1).replace('.', '')

        if version_id is None:
            msg = "Could not find a version specification from the given main file."
            logger.exception(msg)
            raise RuntimeError(msg)

        # Setup the root folder to the fileservice that is used to access files
        dirpath, filename = os.path.split(self.parser_context.main_file)
        dirpath = os.path.abspath(dirpath)
        self.parser_context.file_service.setup_root_folder(dirpath)
        self.parser_context.file_service.set_file_id(filename, "output")

        # Setup the correct main parser based on the version id. If no match
        # for the version is found, use the main parser for NWChem 6.6
        self.setup_main_parser(version_id)

    def get_metainfo_filename(self):
        return "nwchem.nomadmetainfo.json"

    def get_parser_info(self):
        return {'name': 'nwchem-parser', 'version': '1.0'}

    def setup_main_parser(self, version_id):
        # Currently the version id is a pure integer, so it can directly be mapped
        # into a package name.
        base = "nwchemparser.versions.nwchem{}.mainparser".format(version_id)
        parser_module = None
        parser_class = None
        try:
            parser_module = importlib.import_module(base)
        except ImportError:
            logger.warning("Could not find a parser for version '{}'. Trying to default to the base implementation for NWChem 6.6".format(version_id))
            base = "nwchemparser.versions.nwchem66.mainparser"
            try:
                parser_module = importlib.import_module(base)
            except ImportError:
                logger.exception("Tried to default to the NWChem 6.6 implementation but could not find the correct module.")
                raise
        try:
            parser_class = getattr(parser_module, "NWChemMainParser")
        except AttributeError:
            logger.exception("A parser class 'NWChemMainParser' could not be found in the module '[]'.".format(parser_module))
            raise
        self.main_parser = parser_class(self.parser_context)


# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from builtins import object
import setup_paths
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import get_metaInfo
#from TurbomoleCommon import get_metaInfo, write_controlIn, write_k_grid, write_xc_functional
import logging, os, re, sys

############################################################
# This is the parser for the control.in file of FHI-aims.
############################################################

logger = logging.getLogger("nomad.TurbomoleControlInParser")

class TurbomoleControlInParserContext(object):
    """Context for parsing Turbomole control file.

    Attributes:
        sectionRun: Stores the parsed value/sections found in section_run.

    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of writing and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeSectionRun = True):
        """Args:
            writeSectionRun: Deteremines if metadata is written on close of section_run
                or stored in sectionRun.
        """
        self.writeSectionRun = writeSectionRun

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        # save metadata
        self.metaInfoEnv = parser.parserBuilder.metaInfoEnv
        # allows to reset values if the same superContext is used to parse different files
        self.sectionRun = None


def build_TurbomoleControlInKeywordsSimpleMatchers():
    """Builds the list of SimpleMatchers to parse the control keywords of Turbomole.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       List of SimpleMatchers that parses control.in keywords of Turbomole.
    """
    # Now follows the list to match the keywords from the control.in.
    # Explicitly add ^ to ensure that the keyword is not within a comment.
    # Repating occurrences of the same keywords are captured.
    # List the matchers in alphabetical order according to keyword name.
    #
    return [
        SM (r"\s*\$operating system\s+(?P<x_turbomole_controlIn_operating_system>[a-zA-Z\s]+)", repeats = True),
        SM (r"\s*\$symmetry\s+(?P<x_turbomole_controlIn_symmetry>[a-zA-Z0-9])", repeats = True),
        SM (startReStr = r"\s*\$atoms\s+",
            forwardMatch = True,
            #repeats = True,
            subMatchers = [
            SM (r"\s*(?P<x_turbomole_controlIn_atom_label>[a-zA-Z]+)\s+(?P<x_turbomole_controlIn_atom_number>[0-9]+)\s*(?: \\)", repeats = True)#,
            #SM (r"\s*(?P<x_turbomole_controlIn_atom_label>[a-zA-Z]+)\s+(?P<x_turbomole_controlIn_atom_number>[-0-9]+)\s*(?: \\)", repeats = True)
            ]),
        SM (r"\s*\$pople\s+(?P<x_turbomole_controlIn_pople_kind>[a-zA-Z]+)", repeats = True),
        SM (r"\s*\$uhfmo_alpha\s*file\=alpha"),
        SM (r"\s*\$uhfmo_beta\s*file\=beta"),
        SM (r"\s*\$scfiterlimit\s+(?P<x_turbomole_controlIn_scf_iter_limit>[0-9]+)"),
        SM (r"\s*\$scfconv\s+(?P<x_turbomole_controlIn_scf_conv>[0-9]+)"),
        #SM (r"\s*\$thize\s+(?P<x_turbomole_controlIn_number_of_integral_stored>[.0-9.eEdD]+)"),
        SM (r"\s*\$thime\s+(?P<x_turbomole_controlIn_time_for_integral_calc>[0-9]+)"),
        SM (r"\s*\$scfdamp\s+start\=(?P<x_turbomole_controlIn_damping_parameter_start>[.0-9]+)\s*step\=\s*(?P<x_turbomole_controlIn_damping_parameter_step>[.0-9]+)\s*min\=\s*(?P<x_turbomole_controlIn_damping_parameter_min>[.0-9]+)"),
        SM (startReStr = r"\s*\$scfintunit",
            forwardMatch = True,
            subMatchers = [
            SM (r"\s*unit\=(?P<x_turbomole_controlIn_scfintunit_unit>[0-9]+)\s*size\=\s*(?P<x_turbomole_controlIn_scfintunit_size>[0-9]+)\s*file\=\s*(?P<x_turbomole_controlIn_scfintunit_file>[a-zA-Z]+)")#,
            ]),
        SM (startReStr = r"\s*\$drvopt",
            forwardMatch = True,
            subMatchers = [
            SM (r"\s*(?:cartesian)\s+(?P<x_turbomole_controlIn_cartesian_status>[a-zA-Z]+)"),
            SM (r"\s*(?:basis)\s+(?P<x_turbomole_controlIn_basis_status>[a-z-A-Z]+)"),
            SM (r"\s*(?:global)\s+(?P<x_turbomole_controlIn_global_status>[a-zA-Z]+)"),
            SM (r"\s*(?:hessian)\s+(?P<x_turbomole_controlIn_hessian_status>[a-z-A-Z]+)"),
            SM (r"\s*(?:dipole)\s+(?P<x_turbomole_controlIn_dipole_status>[a-zA-Z]+)"),
            SM (r"\s*(?:\$interconversion)\s+(?P<x_turbomole_controlIn_interconversion_status>[a-zA-Z]+)")
            ])
        ]

def build_TurbomoleControlInFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the control.in file of Turbomole.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses control.in file of Turbomole.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
        SM (name = 'Root2',
            startReStr = "",
            sections = ['section_method'],
            forwardMatch = True,
            weak = True,
            # The search is done unordered since the keywords do not appear in a specific order.
            subFlags = SM.SubFlags.Unordered,
            subMatchers = build_TurbomoleControlInKeywordsSimpleMatchers()
            )
        ])

def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections method and run. This allows to run the parser
            without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                               'section_method': CachingLvl,
                               'section_run': CachingLvl,
                              }
    # Set all controlIn metadata to Cache to capture multiple occurrences of keywords and
    # their last value is then written by the onClose routine in the TurbomoleControlInParserContext.
    for name in metaInfoEnv.infoKinds:
        if (name.startswith('x_turbomole_controlIn_') and not name.startswith('x_turbomole_controlIn_basis_')
            and not name.startswith('x_turbomole_controlIn_species_')):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName

def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the Turbomole control.in file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections method and run. This allows to run the parser
            without opening new sections
    """
    # get control.in file description
    TurbomoleControlInSimpleMatcher = build_TurbomoleControlInFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/turbomole.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/turbomole.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'x_turbomole-control-in-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = TurbomoleControlInSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = TurbomoleControlInParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)




# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import typing
import click
import datetime
import elasticsearch_dsl
import elasticsearch
import sys
import threading

from nomad import processing as proc, search, datamodel, infrastructure, utils, config
from nomad.cli.cli import cli


def __run_parallel(
        uploads, parallel: int, callable, label: str):
    if isinstance(uploads, (tuple, list)):
        uploads_count = len(uploads)

    else:
        uploads_count = uploads.count()
        uploads = list(uploads)  # copy the whole mongo query set to avoid cursor timeouts

    cv = threading.Condition()
    threads: typing.List[threading.Thread] = []

    state = dict(
        completed_count=0,
        skipped_count=0,
        available_threads_count=parallel)

    logger = utils.get_logger(__name__)

    print('%d uploads selected, %s ...' % (uploads_count, label))

    def process_upload(upload: proc.Upload):
        logger.info('%s started' % label, upload_id=upload.upload_id)

        completed = False
        if callable(upload, logger):
            completed = True

        with cv:
            state['completed_count'] += 1 if completed else 0
            state['skipped_count'] += 1 if not completed else 0
            state['available_threads_count'] += 1

            print(
                '   %s %s and skipped %s of %s uploads' %
                (label, state['completed_count'], state['skipped_count'], uploads_count))

            cv.notify()

    for upload in uploads:
        with cv:
            cv.wait_for(lambda: state['available_threads_count'] > 0)
            state['available_threads_count'] -= 1
            thread = threading.Thread(target=lambda: process_upload(upload))
            threads.append(thread)
            thread.start()

    for thread in threads:
        thread.join()


def __run_processing(
        uploads, parallel: int, process, label: str, reprocess_running: bool = False):

    def run_process(upload, logger):
        if upload.process_running and not reprocess_running:
            logger.warn(
                'cannot trigger %s, since the upload is already/still processing' % label,
                current_process=upload.current_process,
                current_task=upload.current_task, upload_id=upload.upload_id)
            return False
        else:
            upload.reset()
            process(upload)
            upload.block_until_complete(interval=.5)

            if upload.tasks_status == proc.FAILURE:
                logger.info('%s with failure' % label, upload_id=upload.upload_id)

            logger.info('%s complete' % label, upload_id=upload.upload_id)
            return True

    __run_parallel(uploads, parallel=parallel, callable=run_process, label=label)


@cli.group(help='''The nomad admin commands to do nasty stuff directly on the databases.
                     Remember: With great power comes great responsibility!''')
@click.pass_context
def admin(ctx):
    pass


@admin.command(help='Reset/remove all databases.')
@click.option('--remove', is_flag=True, help='Do not just reset all dbs, but also remove them.')
@click.option('--i-am-really-sure', is_flag=True, help='Must be set for the command to to anything.')
def reset(remove, i_am_really_sure):
    if not i_am_really_sure:
        print('You do not seem to be really sure about what you are doing.')
        sys.exit(1)

    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    infrastructure.reset(remove)


@admin.command(help='Reset all "stuck" in processing uploads and calc in low level mongodb operations.')
@click.option('--zero-complete-time', is_flag=True, help='Sets the complete time to epoch zero.')
def reset_processing(zero_complete_time):
    infrastructure.setup_mongo()

    def reset_collection(cls):
        in_processing = cls.objects(process_status__in=[proc.PROCESS_RUNNING, proc.base.PROCESS_CALLED])
        print('%d %s processes need to be reset due to incomplete process' % (in_processing.count(), cls.__name__))
        in_processing.update(
            process_status=None,
            current_process=None,
            worker_hostname=None,
            celery_task_id=None,
            errors=[], warnings=[],
            complete_time=datetime.datetime.fromtimestamp(0) if zero_complete_time else datetime.datetime.now(),
            current_task=None,
            tasks_status=proc.base.CREATED)

        in_tasks = cls.objects(tasks_status__in=[proc.PENDING, proc.RUNNING])
        print('%d %s processes need to be reset due to incomplete tasks' % (in_tasks.count(), cls.__name__))
        in_tasks.update(
            current_task=None,
            tasks_status=proc.base.CREATED,
            errors=[], warnings=[],
            complete_time=datetime.datetime.fromtimestamp(0) if zero_complete_time else datetime.datetime.now())

    reset_collection(proc.Calc)
    reset_collection(proc.Upload)


@admin.command(help='Check and lift embargo of data with expired embargo period.')
@click.option('--dry', is_flag=True, help='Do not lift the embargo, just show what needs to be done.')
@click.option('--parallel', default=1, type=int, help='Use the given amount of parallel processes. Default is 1.')
def lift_embargo(dry, parallel):
    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    request = search.SearchRequest()
    request.q = elasticsearch_dsl.Q('term', with_embargo=True) & elasticsearch_dsl.Q('term', published=True)
    request.quantity('upload_id', 1000)
    result = request.execute()

    uploads_to_repack = []
    for upload_id in result['quantities']['upload_id']['values']:
        upload = proc.Upload.get(upload_id)
        embargo_length = upload.embargo_length
        if embargo_length is None:
            embargo_length = 36
            upload.embargo_length = 36

        if upload.upload_time + datetime.timedelta(days=int(embargo_length * 365 / 12)) < datetime.datetime.now():
            print('need to lift the embargo of %s (upload_time=%s, embargo=%d)' % (
                upload.upload_id, upload.upload_time, embargo_length))

            if not dry:
                proc.Calc._get_collection().update_many(
                    {'upload_id': upload_id},
                    {'$set': {'metadata.with_embargo': False}})
                uploads_to_repack.append(upload)
                upload.save()

                with upload.entries_metadata() as entries:
                    search.index_all(entries)

    if not dry:
        __run_processing(uploads_to_repack, parallel, lambda upload: upload.re_pack(), 're-packing')


@admin.command(help='(Re-)index all calcs.')
@click.option('--threads', type=int, default=1, help='Number of threads to use.')
@click.option('--dry', is_flag=True, help='Do not index, just compute entries.')
def index(threads, dry):
    infrastructure.setup_mongo()
    infrastructure.setup_elastic()

    all_calcs = proc.Calc.objects().count()
    print('indexing %d ...' % all_calcs)

    def elastic_updates():
        with utils.ETA(all_calcs, '   index %10d or %10d calcs, ETA %s') as eta:
            for calc in proc.Calc.objects():
                eta.add()
                entry_metadata = datamodel.EntryMetadata.m_from_dict(calc.metadata)
                entry = entry_metadata.a_elastic.create_index_entry().to_dict(include_meta=True)
                entry['_op_type'] = 'index'
                yield entry

    if dry:
        for _ in elastic_updates():
            pass
    if threads > 1:
        print('  use %d threads' % threads)
        for _ in elasticsearch.helpers.parallel_bulk(
                infrastructure.elastic_client, elastic_updates(), chunk_size=500,
                thread_count=threads):
            pass
    else:
        elasticsearch.helpers.bulk(
            infrastructure.elastic_client, elastic_updates())
    search.refresh()

    print('')
    print('indexing completed')


@admin.group(help='Generate scripts and commands for nomad operation.')
def ops():
    pass


@ops.group(help='Tools for managing the DOS similarity data.')
def similarity():
    pass


@ops.command(help=('Dump the mongo (calculation metadata) db.'))
@click.option('--restore', is_flag=True, help='Do not dump, but restore.')
def dump(restore: bool):
    date_str = datetime.datetime.utcnow().strftime('%Y_%m_%d')
    print('mongodump --host {} --port {} --db {} -o /backup/fairdi/mongo/{}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, date_str))


@ops.command(help=('Restore the mongo (calculation metadata) db.'))
@click.argument('PATH_TO_DUMP', type=str, nargs=1)
def restore(path_to_dump):
    print('mongorestore --host {} --port {} --db {} {}'.format(
        config.mongo.host, config.mongo.port, config.mongo.db_name, path_to_dump))


@ops.command(help=('Generate an nginx.conf to serve the GUI and proxy pass to API container.'))
@click.option('--prefix', type=str, default='/example_nomad', help='Url path prefix. Default is /example_nomd, can be empty str.')
def nginx_conf(prefix):
    prefix = prefix.rstrip('/')
    prefix = '/%s' % prefix.lstrip('/')

    print('''\
server {{
    listen        80;
    server_name   www.example.com;
    proxy_set_header Host $host;

    location / {{
        proxy_pass http://app:8000;
    }}

    location ~ {1}\\/?(gui)?$ {{
        rewrite ^ {1}/gui/ permanent;
    }}

    location {1}/gui/ {{
        proxy_intercept_errors on;
        error_page 404 = @redirect_to_index;
        proxy_pass http://app:8000;
    }}

    location @redirect_to_index {{
        rewrite ^ {1}/gui/index.html break;
        proxy_pass http://app:8000;
    }}

    location ~ \\/gui\\/(service-worker\\.js|meta\\.json)$ {{
        add_header Last-Modified $date_gmt;
        add_header Cache-Control 'no-store, no-cache, must-revalidate, proxy-revalidate, max-age=0';
        if_modified_since off;
        expires off;
        etag off;
        proxy_pass http://app:8000;
    }}

    location ~ \\/api\\/uploads\\/?$ {{
        client_max_body_size 35g;
        proxy_request_buffering off;
        proxy_pass http://app:8000;
    }}

    location ~ \\/api\\/(raw|archive) {{
        proxy_buffering off;
        proxy_pass http://app:8000;
    }}

    location ~ \\/api\\/mirror {{
        proxy_buffering off;
        proxy_read_timeout 600;
        proxy_pass http://app:8000;
    }}
}}'''.format(prefix))


@ops.command(help=('Generate a proxy pass config for apache2 reverse proxy servers.'))
@click.option('--prefix', type=str, default='app', help='The path prefix under which everything is proxy passed.')
@click.option('--host', type=str, default='130.183.207.104', help='The host to proxy to.')
@click.option('--port', type=str, default='30001', help='The port to proxy to.')
def apache_conf(prefix, host, port):
    print('''\
ProxyPass "/{0}" "http://{1}:{2}/{0}"
ProxyPassReverse "/{0}" "http://{1}:{2}/{0}"
<Proxy http://{1}:{2}/{0}>
    ProxyPreserveHost On
    <IfModule !mod_access_compat.c>
         Require all granted
     </IfModule>
     <IfModule mod_access_compat.c>
         Order allow,deny
         Allow from all
     </IfModule>
</Proxy>

RequestHeader set "X-Forwarded-Proto" expr=%{{REQUEST_SCHEME}}
RequestHeader set "X-Forwarded-SSL" expr=%{{HTTPS}}

ProxyPass /fairdi/keycloak http://{1}:8002/fairdi/keycloak
ProxyPassReverse /fairdi/keycloak http://{1}:8002/fairdi/keycloak
<Proxy http://{1}:8002/app>
     ProxyPreserveHost On
     <IfModule !mod_access_compat.c>
         Require all granted
     </IfModule>
     <IfModule mod_access_compat.c>
         Order allow,deny
         Allow from all
     </IfModule>
</Proxy>

RewriteEngine on
RewriteCond %{QUERY_STRING} ^pid=([^&]+)$
RewriteRule ^/NomadRepository-1.1/views/calculation.zul$ /{0}/gui/entry/pid/%1? [R=301]

AllowEncodedSlashes On
'''.format(prefix, host, port))  # type: ignore


@ops.command(help='Updates the AFLOW prototype information using the latest online version and writes the results to a python module in the given FILEPATH.')
@click.argument('FILEPATH', nargs=1, type=str)
@click.option('--matches-only', is_flag=True, help='Only update the match information that depends on the symmetry analysis settings. Will not perform and online update.')
@click.pass_context
def prototypes_update(ctx, filepath, matches_only):
    from nomad.cli.admin import prototypes
    prototypes.update_prototypes(ctx, filepath, matches_only)


@ops.command(help='Updates the springer database in nomad.config.normalize.springer_db_path.')
@click.option('--max-n-query', default=10, type=int, help='Number of unsuccessful springer request before returning an error. Default is 10.')
@click.option('--retry-time', default=120, type=int, help='Time in seconds to retry after unsuccessful request. Default is 120.')
def springer_update(max_n_query, retry_time):
    from nomad.cli.admin import springer
    springer.update_springer(max_n_query, retry_time)


@similarity.command(help='Updates the msgpack file containing the similarity information.')
@click.option('--dir', "-d", "input_dir", type=str, help='Path of the folder containing the raw similarity information files')
@click.option('--out', "-o", type=str, help='Path of the output msgpack file.')
@click.option('--verbose', is_flag=True, help='Enable verbose output.')
def update(input_dir, out, verbose):
    from nomad.cli.admin import similarity
    similarity.update(input_dir, out, verbose)


@similarity.command(help='Ingests the given similarity information from an msgpack file into MongoDB.')
@click.option('--in', "-i", "input_path", type=str, help='Path of the ingested msgpack file.')
@click.option('--batch_size', type=int, default=10000, help='Batch size for MongoDB bulk ingestion.')
@click.option('--verbose', is_flag=True, help='Enable verbose output.')
def ingest(input_path, batch_size, verbose):
    from nomad.cli.admin import similarity
    similarity.ingest(input_path, batch_size, verbose)



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='elastic_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='elastic.nomadmetainfo.json'))


class x_elastic_section_strain_diagrams(MSection):
    '''
    section collecting the data of the strain diagrams
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elastic_section_strain_diagrams'))

    x_elastic_strain_diagram_values = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations', 'x_elastic_strain_diagram_number_of_eta'],
        description='''
        Values of the energy(units:J)/d2E(units:Pa)/cross-validation (depending on the
        value of x_elastic_strain_diagram_type)
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_strain_diagram_values'))

    x_elastic_strain_diagram_eta_values = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations', 'x_elastic_strain_diagram_number_of_eta'],
        description='''
        eta values used the strain diagrams
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_strain_diagram_eta_values'))

    x_elastic_strain_diagram_number_of_eta = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of strain values used in the strain diagram
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_strain_diagram_number_of_eta'))

    x_elastic_strain_diagram_type = Quantity(
        type=str,
        shape=[],
        description='''
        Kind of strain diagram. Possible values are: energy; cross-validation (cross-
        validation error); d2E (second derivative of the energy wrt the strain)
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_strain_diagram_type'))

    x_elastic_strain_diagram_polynomial_fit_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Order of the polynomial fit
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_strain_diagram_polynomial_fit_order'))


class x_elastic_section_fitting_parameters(MSection):
    '''
    section collecting the fitting parameters used to calculate the elastic constants
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_elastic_section_fitting_parameters'))

    x_elastic_fitting_parameters_eta = Quantity(
        type=np.dtype(np.float64),
        shape=['x_elastic_number_of_deformations'],
        description='''
        eta values used to calculate the elastic constants
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_fitting_parameters_eta'))

    x_elastic_fitting_parameters_polynomial_order = Quantity(
        type=np.dtype(np.int32),
        shape=['x_elastic_number_of_deformations'],
        description='''
        polynomial order used to fit the Energy vs. volume curve and to calculate the
        elastic constants
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_fitting_parameters_polynomial_order'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_elastic_elastic_constant_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Order of the elastic constant
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_elastic_constant_order'))

    x_elastic_number_of_deformations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of deformed structures equally spaced in strain, which are generated
        between the maximum negative strain and the maximum positive one
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_number_of_deformations'))

    x_elastic_deformation_types = Quantity(
        type=np.dtype('U'),
        shape=['x_elastic_number_of_deformations', 6],
        description='''
        deformation types
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_deformation_types'))

    x_elastic_calculation_method = Quantity(
        type=str,
        shape=[],
        description='''
        Method of calculation
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_calculation_method'))

    x_elastic_code = Quantity(
        type=str,
        shape=[],
        description='''
        Code used for the calculation of the elastic constants
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_code'))

    x_elastic_max_lagrangian_strain = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum lagrangian strain used to calculate the elastic constants
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_max_lagrangian_strain'))

    x_elastic_number_of_distorted_structures = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of distorted structures used to calculate the elastic constants
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_number_of_distorted_structures'))

    x_elastic_section_fitting_parameters = SubSection(
        sub_section=SectionProxy('x_elastic_section_fitting_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elastic_section_fitting_parameters'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_elastic_2nd_order_constants_notation_matrix = Quantity(
        type=np.dtype('U'),
        shape=[6, 6],
        description='''
        Symmetry of the second-order elastic constant matrix in Voigt notation
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_notation_matrix'))

    x_elastic_2nd_order_constants_matrix = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        unit='pascal',
        description='''
        2nd order elastic constant (stiffness) matrix in GPa
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_matrix'))

    x_elastic_3rd_order_constants_matrix = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6, 6],
        unit='pascal',
        description='''
        3rd order elastic constant (stiffness) matrix in GPa
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_3rd_order_constants_matrix'))

    x_elastic_2nd_order_constants_compliance_matrix = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        unit='1 / pascal',
        description='''
        Elastic compliance matrix in 1/GPa
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_2nd_order_constants_compliance_matrix'))

    x_elastic_Voigt_bulk_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt bulk modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Voigt_bulk_modulus'))

    x_elastic_Voigt_shear_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt shear modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Voigt_shear_modulus'))

    x_elastic_Reuss_bulk_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss bulk modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Reuss_bulk_modulus'))

    x_elastic_Reuss_shear_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss shear modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Reuss_shear_modulus'))

    x_elastic_Hill_bulk_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill bulk modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Hill_bulk_modulus'))

    x_elastic_Hill_shear_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill shear modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Hill_shear_modulus'))

    x_elastic_Voigt_Young_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Voigt Young modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Voigt_Young_modulus'))

    x_elastic_Voigt_Poisson_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Voigt Poisson ratio
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Voigt_Poisson_ratio'))

    x_elastic_Reuss_Young_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Reuss Young modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Reuss_Young_modulus'))

    x_elastic_Reuss_Poisson_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Reuss Poisson ratio
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Reuss_Poisson_ratio'))

    x_elastic_Hill_Young_modulus = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        Hill Young modulus
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Hill_Young_modulus'))

    x_elastic_Hill_Poisson_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hill Poisson ratio
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_Hill_Poisson_ratio'))

    x_elastic_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        unit='pascal',
        description='''
        Eigemvalues of the stiffness matrix
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_eigenvalues'))

    x_elastic_section_strain_diagrams = SubSection(
        sub_section=SectionProxy('x_elastic_section_strain_diagrams'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_elastic_section_strain_diagrams'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_elastic_space_group_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Space-group number of the system
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_space_group_number'))

    x_elastic_unit_cell_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Volume of the equilibrium unit cell
        ''',
        a_legacy=LegacyDefinition(name='x_elastic_unit_cell_volume'))


m_package.__init_metainfo__()



# pylint: disable=import-outside-toplevel,no-name-in-module
import abc
from typing import Dict

from pydantic import BaseModel

from fastapi.testclient import TestClient


def get_regular_client() -> TestClient:
    """Return TestClient for regular OPTIMADE server"""
    from optimade.server.main import app
    from optimade.server.routers import info, links, references, structures

    # We need to remove the version prefixes in order to have the tests run correctly.
    app.include_router(info.router)
    app.include_router(links.router)
    app.include_router(references.router)
    app.include_router(structures.router)
    # need to explicitly set base_url, as the default "http://testserver"
    # does not validate as pydantic AnyUrl model
    return TestClient(app, base_url="http://example.org/v0")


def get_index_client() -> TestClient:
    """Return TestClient for index meta-database OPTIMADE server"""
    from optimade.server.main_index import app
    from optimade.server.routers import index_info, links

    # # We need to remove the version prefixes in order to have the tests run correctly.
    app.include_router(index_info.router)
    app.include_router(links.router)
    # need to explicitly set base_url, as the default "http://testserver"
    # does not validate as pydantic UrlStr model
    return TestClient(app, base_url="http://example.org/v0")


class SetClient(abc.ABC):
    """Metaclass to instantiate the TestClients once"""

    server: str = None
    _client: Dict[str, TestClient] = {
        "index": get_index_client(),
        "regular": get_regular_client(),
    }

    @property
    def client(self) -> TestClient:
        exception_message = "Test classes using EndpointTestsMixin MUST specify a `server` attribute with a value that is either 'regular' or 'index'"
        if not hasattr(self, "server"):
            raise AttributeError(exception_message)
        if self.server in self._client:
            return self._client[self.server]
        raise ValueError(exception_message)

    # pylint: disable=no-member
    def _check_error_response(
        self,
        request: str,
        expected_status: int = None,
        expected_title: str = None,
        expected_detail: str = None,
    ):
        try:
            response = self.client.get(request)
            self.assertEqual(
                response.status_code,
                expected_status,
                msg=f"Request should have been an error with status code {expected_status}, "
                f"but instead {response.status_code} was received.\nResponse:\n{response.json()}",
            )
            response = response.json()
            self.assertEqual(len(response["errors"]), 1)
            self.assertEqual(response["meta"]["data_returned"], 0)

            error = response["errors"][0]
            self.assertEqual(str(expected_status), error["status"])
            self.assertEqual(expected_title, error["title"])

            if expected_detail is None:
                expected_detail = "Error trying to process rule "
                self.assertTrue(error["detail"].startswith(expected_detail))
            else:
                self.assertEqual(expected_detail, error["detail"])

        except Exception as exc:
            print("Request attempted:")
            print(f"{self.client.base_url}{request}")
            raise exc


class EndpointTestsMixin(SetClient):
    """ Mixin "base" class for common tests between endpoints. """

    server: str = "regular"
    request_str: str = None
    response_cls: BaseModel = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.response = self.client.get(self.request_str)
        self.json_response = self.response.json()
        self.assertEqual(
            self.response.status_code,
            200,
            msg=f"Request failed: {self.response.json()}",
        )

    def test_meta_response(self):
        self.assertTrue("meta" in self.json_response)
        meta_required_keys = [
            "query",
            "api_version",
            "time_stamp",
            "data_returned",
            "more_data_available",
            "provider",
        ]
        meta_optional_keys = ["data_available", "implementation"]

        self.check_keys(meta_required_keys, self.json_response["meta"])
        self.check_keys(meta_optional_keys, self.json_response["meta"])

    def test_serialize_response(self):
        self.assertTrue(
            self.response_cls is not None, msg="Response class unset for this endpoint"
        )
        self.response_cls(**self.json_response)  # pylint: disable=not-callable

    def check_keys(self, keys: list, response_subset: dict):
        for key in keys:
            self.assertTrue(
                key in response_subset,
                msg="{} missing from response {}".format(key, response_subset),
            )



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
from .commonparser import CPMDCommonParser
import re
import logging
import numpy as np
LOGGER = logging.getLogger("nomad")


class CPMDSinglePointParser(MainHierarchicalParser):
    """The main parser class that is called for all run types. Parses the CPMD
    output file.
    """
    def __init__(self, parser_context):
        """
        """
        super(CPMDSinglePointParser, self).__init__(parser_context)
        self.setup_common_matcher(CPMDCommonParser(parser_context))
        self.n_scf_iterations = 0

        #=======================================================================
        # Cache levels
        # self.caching_levels.update({
            # 'section_run': CachingLevel.ForwardAndCache,
        # })

        #=======================================================================
        # SimpleMatchers
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run', "section_single_configuration_calculation", "section_system", "section_method"],
            subMatchers=[
                self.cm.header(),
                self.cm.method(),
                self.cm.atoms(),
                self.cm.cell(),
                self.cm.initialization(),
                SM( " NFI      GEMAX       CNORM           ETOT        DETOT      TCPU",
                    sections=["x_cpmd_section_scf"],
                    subMatchers=[
                        SM( "\s+(?P<x_cpmd_scf_nfi>{0})\s+(?P<x_cpmd_scf_gemax>{1})\s+(?P<x_cpmd_scf_cnorm>{1})\s+(?P<x_cpmd_scf_etot__hartree>{1})\s+(?P<x_cpmd_scf_detot__hartree>{1})\s+(?P<x_cpmd_scf_tcpu__s>{1})".format(self.regexs.int, self.regexs.float),
                            sections=["x_cpmd_section_scf_iteration"],
                            repeats=True,
                        ),
                    ]
                ),
                SM( re.escape(" *                        FINAL RESULTS                         *"),
                    sections=["x_cpmd_section_final_results"],
                    subMatchers=[
                        SM( "   ATOM          COORDINATES            GRADIENTS \(-FORCES\)",
                            adHoc=self.ad_hoc_atom_forces(),
                        ),
                        SM( " \(K\+E1\+L\+N\+X\)           TOTAL ENERGY =\s+(?P<energy_total__hartree>{}) A\.U\.".format(self.regexs.float)),
                        SM( " \(E1=A-S\+R\)     ELECTROSTATIC ENERGY =\s+(?P<energy_electrostatic__hartree>{}) A\.U\.".format(self.regexs.float)),
                        SM( " \(X\)     EXCHANGE-CORRELATION ENERGY =\s+(?P<energy_XC_potential__hartree>{}) A\.U\.".format(self.regexs.float)),
                    ]
                ),
                self.cm.footer(),
            ]
        )

    #=======================================================================
    # onClose triggers
    def onClose_x_cpmd_section_scf_iteration(self, backend, gIndex, section):
        # SCF step energy and energy change
        scf_id = backend.openSection("section_scf_iteration")
        energy = section.get_latest_value("x_cpmd_scf_etot")
        backend.addValue("energy_total_scf_iteration", energy)
        denergy = section.get_latest_value("x_cpmd_scf_detot")
        backend.addValue("energy_change_scf_iteration", denergy)
        backend.closeSection("section_scf_iteration", scf_id)
        self.n_scf_iterations += 1

    def onClose_x_cpmd_section_scf(self, backend, gIndex, section):
        backend.addValue("number_of_scf_iterations", self.n_scf_iterations)

    def onClose_section_system(self, backend, gIndex, section):
        self.cache_service.addArrayValues("atom_positions", "initial_positions", unit="bohr")
        self.cache_service.addArrayValues("atom_labels")
        self.cache_service.addArrayValues("simulation_cell", unit="bohr")
        self.cache_service.addValue("number_of_atoms")

    #=======================================================================
    # adHoc

    def ad_hoc_atom_forces(self):
        """Parses the atomic forces from the final results.
        """
        def wrapper(parser):
            # Define the regex that extracts the information
            regex_string = r"\s+({0})\s+({1})\s+({2})\s+({2})\s+({2})\s+({2})\s+({2})\s+({2})".format(self.regexs.int, self.regexs.word, self.regexs.float)
            regex_compiled = re.compile(regex_string)

            match = True
            forces = []

            while match:
                line = parser.fIn.readline()
                result = regex_compiled.match(line)

                if result:
                    match = True
                    force = [float(x) for x in result.groups()[5:8]]
                    forces.append(force)
                else:
                    match = False

            forces = -np.array(forces)

            # If anything found, push the results to the correct section
            if len(forces) != 0:
                parser.backend.addArrayValues("atom_forces", forces, unit="hartree/bohr")

        return wrapper

    #=======================================================================
    # Misc. functions
